import enum


class ObjectType(enum.StrEnum):
    NONE = "NONE"
    ANY = "ANY"
    EVENT = "EVENT"
    MDS = "MDS"
    NODE = "NODE"
    FAN = "FAN"
    NIC = "NIC"
    HDD = "HDD"
    SSD = "SSD"
    DISK_DRIVE = "DISK_DRIVE"
    POWER_SUPPLY = "POWER_SUPPLY"
    DATA_SPHERE = "DATA_SPHERE"
    DATA_MOVER = "DATA_MOVER"
    DATA_PORTAL = "DATA_PORTAL"
    STORAGE_VOLUME = "STORAGE_VOLUME"
    VOLUME_GROUP = "VOLUME_GROUP"
    NETWORK_IF = "NETWORK_IF"
    MGMT = "MGMT"
    SHARE = "SHARE"
    BASE_USER = "BASE_USER"
    USER = "USER"
    FEDERATED_USER = "FEDERATED_USER"
    USER_GROUP = "USER_GROUP"
    IDENTITY_GROUP_MAPPING = "IDENTITY_GROUP_MAPPING"
    IDENTITY = "IDENTITY"
    MANAGEMENT_ROLE = "MANAGEMENT_ROLE"
    STEP = "STEP"
    TASK = "TASK"
    SMTP = "SMTP"
    NOTIFICATION_RULE = "NOTIFICATION_RULE"
    HEARTBEAT = "HEARTBEAT"
    NTP = "NTP"
    LOGICAL_VOLUME = "LOGICAL_VOLUME"
    LDAP = "LDAP"
    IDP = "IDP"
    DNS = "DNS"
    GATEWAY = "GATEWAY"
    DHCP = "DHCP"
    CLUSTER = "CLUSTER"
    CPU = "CPU"
    ELEMENTAL_OBJECTIVE = "ELEMENTAL_OBJECTIVE"
    OBJECTIVE = "OBJECTIVE"
    DATA_MANAGER = "DATA_MANAGER"
    SCHEDULE = "SCHEDULE"
    CONDITION = "CONDITION"
    SNAPSHOT_RETENTION = "SNAPSHOT_RETENTION"
    FILE_SNAPSHOT = "FILE_SNAPSHOT"
    OBJECT_STORE = "OBJECT_STORE"
    OBJECT_STORAGE_VOLUME = "OBJECT_STORAGE_VOLUME"
    CLOUD_MOVER = "CLOUD_MOVER"
    OBJECT_STORE_LOGICAL_VOLUME = "OBJECT_STORE_LOGICAL_VOLUME"
    NODE_LOCATION = "NODE_LOCATION"
    VOLUME_LOCATION = "VOLUME_LOCATION"
    SNMP = "SNMP"
    SW_UPDATE_TASK = "SW_UPDATE_TASK"
    CTDB = "CTDB"
    SHARE_SNAPSHOT = "SHARE_SNAPSHOT"
    BACKUP = "BACKUP"
    SAMBA_AD = "SAMBA_AD"
    DOMAIN_IDMAP = "DOMAIN_IDMAP"
    HMDB = "HMDB"
    SYSLOG = "SYSLOG"
    NIS = "NIS"
    NIS_CONFIG = "NIS_CONFIG"
    SUBNET_GATEWAY = "SUBNET_GATEWAY"
    STATIC_ROUTE = "STATIC_ROUTE"
    SITE = "SITE"
    SHARE_PARTICIPANT = "SHARE_PARTICIPANT"
    LOGIN_POLICY = "LOGIN_POLICY"
    DATA_COPY_TO_OBJECT_TASK = "DATA_COPY_TO_OBJECT_TASK"
    KMS = "KMS"
    BASE_STORAGE_VOLUME = "BASE_STORAGE_VOLUME"
    LICENSE_ACTIVATION = "LICENSE_ACTIVATION"
    SMB_ALIAS = "SMB_ALIAS"
    LOGICAL_SHARE = "LOGICAL_SHARE"
    WORM_ENTRY = "WORM_ENTRY"
    ANTIVIRUS = "ANTIVIRUS"
    ALCHEMI = "ALCHEMI"
    LABEL = "LABEL"
    S3_SERVER_CONFIG = "S3_SERVER_CONFIG"
    S3_SERVER_USER = "S3_SERVER_USER"
    S3_SERVER_BUCKET = "S3_SERVER_BUCKET"
    S3_SERVER_BUCKET_CONTAINER = "S3_SERVER_BUCKET_CONTAINER"
